from .ontbo import Ontbo
from .profile import Profile
from .query_type import QueryType
from .scene_message import SceneMessage
from .scene import Scene
from .update_status import UpdateStatus
from .exceptions import ProfileNotFoundError, SceneNotFoundError



import os
from pathlib import Path
from dotenv import load_dotenv

# Resolve the path to tests/.env.test
_TESTS_DIR = Path(__file__).parent
_ENV_PATH = _TESTS_DIR / ".env.test"

# Load variables into process env before any tests run
load_dotenv(dotenv_path=_ENV_PATH, override=True)

# (optional) sanity check, can be removed later
if not os.getenv("API_KEY"):
    raise RuntimeError(f"Failed to load {_ENV_PATH}")




import os
from ontbo import Ontbo, SceneMessage, ProfileNotFoundError, SceneNotFoundError, UpdateStatus

API_KEY=os.getenv("API_KEY")
ONTBO_SERVER_ROOT=os.getenv("ONTBO_SERVER_ROOT")

ontbo = Ontbo(API_KEY, base_url=ONTBO_SERVER_ROOT)


# Scene.id

# Scene.exists

def test_scene_exists():
    ontbo = Ontbo(API_KEY, ONTBO_SERVER_ROOT)

    profile = ontbo.create_profile()

    scene = profile.scene("this_scene_does_not_exists")

    assert not scene.exists

    scene = profile.create_scene()

    assert scene.exists

    ontbo.delete_profile(profile.id)

# Scene.title

def test_scene_title():
    ontbo = Ontbo(API_KEY, ONTBO_SERVER_ROOT)

    profile = ontbo.create_profile()

    scene = profile.create_scene(title="Initial Title")

    assert scene.title == "Initial Title"

    scene.set_properties(title="Updated Title")

    updated_scene = profile.scene(scene.id)

    assert updated_scene.title == "Updated Title"

    ontbo.delete_profile(profile.id)

# Scene.source_type
def test_scene_source_type():
    ontbo = Ontbo(API_KEY, ONTBO_SERVER_ROOT)

    profile = ontbo.create_profile()

    scene = profile.create_scene(source_type="chatgpt")

    assert scene.source_type == "chatgpt"

    scene.set_properties(source_type="claude")

    updated_scene = profile.scene(scene.id)

    assert updated_scene.source_type == "claude"

    ontbo.delete_profile(profile.id)

# Scene.source_url

# Scene.set_properties

# Scene.add_messages

# Scene.clear_messages

# Scene.messages


from ontbo import Ontbo, SceneMessage
from pathlib import Path
import json


if __name__ == "__main__":

    # Script params.
    API_KEY="<YOUR_API_KEY>"
    PROFILE_NAME = "youssef"
    DATA_PATH = "datasets/youssef_50_sessions"
    
    ontbo = Ontbo(API_KEY)

    # (optional) If a profile with the same ID already exists, we delete it.
    if PROFILE_NAME in ontbo.profile_ids:
        ontbo.delete_profile(PROFILE_NAME)

    # Create the new profile.
    profile=ontbo.create_profile(PROFILE_NAME)

    # Check the ID of the created profile.
    print(f"Profile created profile ID is {profile.id}.")
    print(f"Use this ID to query the profile in the query script")

    # Get the list of JSON files to upload.
    data_dir = Path(__file__).resolve().parent / DATA_PATH
    json_files = list(data_dir.glob("*.json"))

    chat_count = len(json_files)
    current_chat=0

    for file in json_files:
        current_chat += 1
        print(f"Uploading chat {current_chat}/{chat_count}")

        # Load JSON data, each message is a dict, with the following structure:
        # {
        #   "content":content,
        #   "role":role,
        #   "timestamp":timestamp,
        # }
        with open(file) as f:
            conversation_data = json.loads(f.read())

        # Each dict in the conversation is conerted in a SceneMessage.
        mesages = [SceneMessage(
                content=message["content"],
                role=message["role"],
                timestamp=message["timestamp"]
            ) for message in conversation_data]
        
        # Create the scene and upload the messages.
        # If you want to be sure that the profile computation is done when the 
        # script ends, set wait_for_result to true.
        scene=profile.create_scene("scene")
        scene.add_messages(mesages, update_now=True, wait_for_result=False)

    print(f"Scene upload complete, whait for a few minutes so the server"\
          "processes all scenes.")
    





import os
import time
from ontbo import Ontbo, SceneMessage, ProfileNotFoundError, SceneNotFoundError, UpdateStatus

API_KEY=os.getenv("API_KEY")
ONTBO_SERVER_ROOT=os.getenv("ONTBO_SERVER_ROOT")

ontbo = Ontbo(API_KEY, base_url=ONTBO_SERVER_ROOT)

PROFILE_UPDATE_MAX_SECONDS_WAIT = 30

###############################################################################
# Tests for the Profile class
###############################################################################


# Profile.exists

def test_exists():

    profile = ontbo.create_profile()
    assert profile.exists

    ontbo.delete_profile(profile.id)

    assert not profile.exists

# Profile.scene_ids

def test_scene_ids():
    SCENES_COUNT = 10

    profile = ontbo.create_profile()

    created_scenes_ids = []

    for i in range(SCENES_COUNT):
        created_scenes_ids.append(profile.create_scene().id)


    profile_scenes = profile.scene_ids

    for id in created_scenes_ids:
        assert id in profile_scenes

    for id in profile_scenes:
        assert id in created_scenes_ids

    ontbo.delete_profile(profile.id)

def test_scene_ids_on_non_existing_profile():
    
    profile_id = "this_profile_should_not_exist_"
    suffix_id = 0
    existing_ids = ontbo.profile_ids

    while f"{profile_id}{suffix_id}" in existing_ids:
        suffix_id += 1

    exception_raised = False

    try:
        ids = ontbo.profile(f"{profile_id}{suffix_id}").scene_ids
    except ProfileNotFoundError as e:
        exception_raised = True

    assert exception_raised

# Profile.scene

def test_scene_accessor():
    profile_id = "some_random_profile"
    scene_id = "some_random_scene"

    scene = ontbo.profile("some_random_profile").scene("some_random_scene")

    assert scene.id == scene_id
    assert scene._profile_id == profile_id

# Profile.create_scene


def test_create_scene():
    profile = ontbo.create_profile()

    assert profile.scene_ids == []

    new_scene = profile.create_scene()

    assert profile.scene_ids == [new_scene.id]

    ontbo.delete_profile(profile.id)

def test_create_scene_with_params():
    SCENE_TITLE = "This is some random scene title"
    SCENE_SOURCE_URL = "https://my-random-url.com/here"
    SCENE_SOURCE_TYPE = "somesource:yay"
    SCENE_PREFIX = "sceneprefix"

    profile = ontbo.create_profile()

    assert profile.scene_ids == []

    new_scene = profile.create_scene(
        title=SCENE_TITLE,
        source_url=SCENE_SOURCE_URL,
        source_type=SCENE_SOURCE_TYPE,
        prefix=SCENE_PREFIX
    )

    assert profile.scene_ids == [new_scene.id]
    assert new_scene.title == SCENE_TITLE
    assert new_scene.source_url == SCENE_SOURCE_URL
    assert new_scene.source_type == SCENE_SOURCE_TYPE
    print(new_scene.id)
    assert new_scene.id.startswith(SCENE_PREFIX)

    ontbo.delete_profile(profile.id)       

# Profile.delete_scene


def test_delete_scene():
    profile = ontbo.create_profile()

    assert profile.scene_ids == []

    existing_scenes = []
    for i in range(5):
        existing_scenes.append(profile.create_scene())

    scene_list_before = profile.scene_ids
    
    new_scene = profile.create_scene()

    assert new_scene.id in profile.scene_ids
    assert new_scene.id not in scene_list_before

    profile.delete_scene(new_scene.id)
    assert new_scene.id not in profile.scene_ids

    ontbo.delete_profile(profile.id)

def test_delete_scene_non_existing():
    exception_raised = False
    try:
        ontbo.profile("Non_existing_profile").delete_scene("some_scene")

    except ProfileNotFoundError:
        exception_raised = True
    
    assert exception_raised

    exception_raised = False

    profile = ontbo.create_profile()
    try:
        profile.delete_scene("some_scene")

    except SceneNotFoundError:
        exception_raised = True
    
    assert exception_raised

    ontbo.delete_profile(profile.id)    


# Profile.update

def test_profile_update():
    profile = ontbo.create_profile()

    scene = profile.create_scene()

    scene.add_messages(
        [SceneMessage("Hi, my name is Mike !")]
        )
    
    status = profile.update()
    assert status.pending == 1

    ontbo.delete_profile(profile.id)


def test_profile_update_on_non_existing_profile():
    profile = ontbo.profile("this_profile_does_not_exist")

    assert not profile.exists

    exception_raised = False

    try:
        profile.update()
    except ProfileNotFoundError:
        exception_raised = True

    assert exception_raised

# Profile.update_status

def test_update_status():
    profile = ontbo.create_profile()

    # This profile has just been created, so the pending sount should be zero.
    assert profile.update_status().pending == 0

    scene = profile.create_scene()

    scene.add_messages(
        [SceneMessage("Hi, my name is Mike !")]
        )
    
    status = profile.update()
    assert status.pending == profile.update_status().pending
    
    ontbo.delete_profile(profile.id)
    
def test_update_status_on_non_existing():
    profile = ontbo.profile("non_existing_profile")

    exception_raised = False

    try:
        profile.update_status()
    except ProfileNotFoundError:
        exception_raised = True

    assert exception_raised

# Profile.query_facts
# Profile.build_context

def test_profile_query_facts():

    profile = ontbo.create_profile()

    # This profile has just been created, so the pending sount should be zero.
    assert profile.update_status().pending == 0

    scene = profile.create_scene()

    scene.add_messages(
        [SceneMessage("Hi, my name is Mike and I am a software engineer !")]
        )
    
    status = profile.update()
    assert status.pending == profile.update_status().pending

    retry_count = PROFILE_UPDATE_MAX_SECONDS_WAIT

    while profile.update_status().pending != 0 and retry_count > 0:
        time.sleep(1)

    assert retry_count > 0

    assert profile.query_facts("What is the user's name?").find("Mike")
    assert profile.query_facts("What is the user's job?").find("engineer")

    assert profile.build_context("White me an introduction letter for a job").find("engineer")

    ontbo.delete_profile(profile.id)    

# Profile.append_facts

def test_profile_append_facts():


    profile = ontbo.create_profile()

    # This profile has just been created, so the pending sount should be zero.
    assert profile.update_status().pending == 0

    profile.append_facts("I like bananas")

    profile.update()

    retry_count = PROFILE_UPDATE_MAX_SECONDS_WAIT

    while profile.update_status().pending != 0 and retry_count > 0:
        time.sleep(1)

    assert retry_count > 0

    assert profile.query_facts("What fruit does the user like?").find("Banana")

    ontbo.delete_profile(profile.id)    


# Profile.list_facts
# Profile.get_fact_details
# Profile.delete_fact

def test_profile_list_facts():

    profile = ontbo.create_profile()

    assert profile.list_facts() == []

    profile.append_facts("I like bananas")

    profile.update()

    while profile.update_status().pending != 0 and retry_count > 0:
        time.sleep(1)

    retry_count = PROFILE_UPDATE_MAX_SECONDS_WAIT

    assert retry_count > 0

    facts = profile.list_facts(fields=["id", "data", "timestamp"])

    assert facts != []

    assert "id" in facts[0].keys()
    assert "data" in facts[0].keys()
    assert "timestamp" in facts[0].keys()

    fact_id = facts[0]["id"]


    profile.delete_fact(fact_id)

    facts = profile.list_facts(fields=["id", "data", "timestamp"])

    assert fact_id not in [fact["id"] for fact in facts]


    ontbo.delete_profile(profile.id)    




# TODO when implemented
# Profile.find_in_scenes





import os
from ontbo import Ontbo, ProfileNotFoundError

API_KEY=os.getenv("API_KEY")
ONTBO_SERVER_ROOT=os.getenv("ONTBO_SERVER_ROOT")

ontbo = Ontbo(API_KEY, base_url=ONTBO_SERVER_ROOT)


###############################################################################
# Tests for the Ontbo class
###############################################################################


def test_incorrect_api_key():
    ontbo_wrong_key = Ontbo("a_very_suspiscious_api_key_f465fds645fdsqf645qsdc645cqcdsq", base_url=ONTBO_SERVER_ROOT)

    excepted = False
    try:
        ids = ontbo_wrong_key.profile_ids()
    except Exception as e:
        excepted = True

    assert excepted  

# Ontbo.profile_ids

def test_list_profiles():
    
    profiles = ontbo.profile_ids

    assert type(profiles) == list

    for profile_id in profiles:
        assert type(profile_id) == str

# Ontbo.profile

def test_profile():
    profile_id = "some_random_id"

    profile = ontbo.profile(profile_id)

    assert profile.id == profile_id

# Ontbo.create_profile

def test_create_profile():
    desired_id = "some_random_profile_id"
    old_profiles = ontbo.profile_ids
    profile = ontbo.create_profile(desired_id)

    assert profile.id not in old_profiles
    assert profile.id in ontbo.profile_ids

    # We delete the profile to keep everything clean

    ontbo.delete_profile(profile.id)

def test_create_profile_without_id():

    old_profiles = ontbo.profile_ids
    profile = ontbo.create_profile()

    assert profile.id not in old_profiles
    assert profile.id in ontbo.profile_ids

    ontbo.delete_profile(profile.id)

# Ontbo.delete_profile

def test_delete_profile():

    profiles_start = ontbo.profile_ids
    profile = ontbo.create_profile()

    assert profile.id not in profiles_start
    assert profile.id in ontbo.profile_ids

    ontbo.delete_profile(profile.id)

    assert profile.id not in ontbo.profile_ids


def test_delete_profile_nonexistent():
    profile_id = "this_profile_should_not_exist_"
    suffix_id = 0
    existing_ids = ontbo.profile_ids

    while f"{profile_id}{suffix_id}" in existing_ids:
        suffix_id += 1

    exception_raised = False

    try:
        ontbo.delete_profile(f"{profile_id}{suffix_id}")
    except ProfileNotFoundError as e:
        exception_raised = True

    assert exception_raised
        




class ProfileNotFoundError(LookupError):
    def __init__(self, profile_id, details=None):
        self._profile_id = profile_id
        self._details = details

    def __str__(self):
        details_str = f"({self._details})" if self._details is not None else ""
        return f"Profile {self._profile_id} not found {details_str}."


class SceneNotFoundError(LookupError):
    def __init__(self, scene_id, details=None):
        self._scene_id = scene_id
        self._details = details

    def __str__(self):
        details_str = f"({self._details})" if self._details is not None else ""
        return f"Scene {self._scene_id} not found {details_str}."

class InvalidResponseError(Exception):
    def __init__(self, details=None):
        self._details = details

    def __str__(self):
        details_str = f": ({self._details})" if self._details is not None else "."
        return f"The server returned an invalid response{details_str}"


from enum import Enum


class QueryType(Enum):
    """
    QueryTypes are used as a parameter query methods to select the algorithm 
    to use to retrieve the relevant information in the user profile.
    """

    FULL_DATA = "FULL_DATA"
    """
    With FULL_DATA query, the algorithm makes a review of ALL items present in 
    the user profile. This method yields the best performance in terms of 
    recall rate, but with a higher token cost. 
    """

    MULTI_HOP = "MULTI_HOP"
    """
    When the query type is set to MULTI_HOP, the agent autonomously tries 
    different strategies and queries to find the relevant information in the 
    profile. This method is fit when the information to find about the user
    is not obvous to find, with a high latency as a counterpart.
    """

    SINGLE_HOP = "SINGLE_HOP"
    """
    Uses a low-latency vector search to find relvant information about the
    profile, then formats the search result using a LLM. Offers good recall
    rate and low latency.
    """

    VECTOR_SEARCH = "VECTOR_SEARCH"
    """
    Directly returns the result of a vector (embeddings) search in the user
    profile database. The best-matching results of the vector quezry are
    returned, but some irrelvant results might also be returned, requiring
    a post-processing by the host application.
    Offers very low letency results.
    """



from urllib.parse import urljoin
from typing import List

from ontbo.i_ontbo_server import IOntboServer
from ontbo.scene_message import SceneMessage

import json
import requests


class Scene:
    """
    A scene is a unit of interaction between the Profile (user) and the system.

    You can get an existing scene by calling the method:
        Ontbo(api_key).profile(profile_id).scene(id)

    Or, you can also create a new scene for a profile by calling the method:
        Ontbo(api_key).profile(profile_id).create_scene(id)
    """

    def __init__(self, server: IOntboServer, profile_id: str, id: str):
        """
        Initialize a Scene instance.

        Args:
            server (IOntboServer): The Ontbo server connection.
            profile_id (str): The ID of the profile associated with this scene.
            id (str): The unique ID of the scene.
        """
        self._server = server
        self._profile_id = profile_id
        self._id = id
        self._data = None


    def _lazy_get(self, property_id: str, force_refresh: bool = False):
        """Utility method to retrieve the value of a property without loading
        data from the server at instantiation.
        """
        if self._data is None or force_refresh:
            response = requests.get(
                    urljoin(self._server.url, f"profiles/{self._profile_id}/scenes/{self._id}"),
                    headers=self._server.headers,
                )
            
            if response.status_code == 404:
                raise LookupError(f"Scene {self._id} doesn't exist for profile {self._profile_id}")
            
            self._data = response.json()

        try:
            return self._data[property_id]
        except KeyError:
            raise KeyError(f"Property '{property_id}' not found in scene data for scene {self._id}")


    @property
    def id(self) -> str:
        """
        Get the scene unique id.

        Returns:
            (str)The unique ID of the scene.
        """
        return self._id
    
    @property
    def exists(self) -> bool:
        """Checks on the server if the profile actually exists."""

        response = requests.get(
            urljoin(self._server.url, f"profiles/{self._profile_id}/scenes/{self._id}"),
            headers=self._server.headers,
        )

        if response.status_code == 404:
            return False
        else:
            self._data = response.json()
            return True    
        
    @property
    def title(self) -> str:
        return self._lazy_get('title')

    @property
    def source_type(self) -> str:
        return self._lazy_get('source_type')

    @property
    def source_url(self) -> str:
        return self._lazy_get('source_url')     

    def set_properties(self,
            title: str|None = None,
            source_type: str|None = None,
            source_url: str|None = None):
        """Updates the properties of a scene after creation. If a property
        value is set to None, it is not updated. 

        Args:
            title (str, optional): the user-friendly title of the scene.

            source_type (str, optional): host-specific string to identify the 
            type of source used for the import (app id, for example)

            source_url (str, optional): the url of the resource used to create
            this scene.

        """

        req_params = {}

        self._lazy_get('title')

        if title is not None:
            self._data['title'] = title
            req_params['title'] = title

        if source_type is not None:
            self._data['source_type'] = source_type
            req_params['source_type'] = source_type

        if source_url is not None:
            self._data['source_url'] = source_url
            req_params['source_url'] = source_url

        response = requests.put(
            urljoin(self._server.url, f"profiles/{self._profile_id}/scenes/{self._id}"),
            params=req_params,
            headers=self._server.headers
        )
        response.raise_for_status()
        
    def add_messages(
        self,
        messages: List[SceneMessage],
        update_now: bool = False
    ) -> str:
        """
        Add messages to the scene.

        Args:
            messages (List[SceneMessage]): a list of SceneMessage objects to
            add to the scene.
            update_now (bool): If set to true, the profile update is initiated 
            now. If set to false, profile might be updated later with
            other calls to Scene.add_messages(), or with a call to 
            Profile.update()

        Returns:
            str: The ID of the newly added message batch.
        """
        text_data = [message.as_dict for message in messages]

        response = requests.post(
            urljoin(self._server.url,
                    f"profiles/{self._profile_id}/scenes/{self._id}/text"),
            json=text_data,
            params={
                "update_now": update_now
            },
            headers=self._server.headers
            )
        response.raise_for_status()
        return response.json()["id"]

    def clear_messages(self) -> None:
        """
        Clears all messages in the scene. 
        """
        response = requests.delete(
            urljoin(self._server.url,
                    f"profiles/{self._profile_id}/scenes/{self._id}/text"),
            headers=self._server.headers,
        )
        response.raise_for_status()

    @property
    def messages(self) -> List[SceneMessage]:
        """
        Returns:
            List[SCeneMessage]: the conversational data of the scene.
        """
        response = requests.get(
            urljoin(self._server.url,
                    f"profiles/{self._profile_id}/scenes/{self._id}/text"),
            headers=self._server.headers,
        )
        response.raise_for_status()
        messages = response.json()

        return [
            SceneMessage(
                content=message["content"],
                role=message["role"],
                timestamp=message["timestamp"],
            )
            for message in messages
        ]



from ontbo import Ontbo, SceneMessage
from pathlib import Path
import json


if __name__ == "__main__":
    # The ID of your newly created profile. 
    API_KEY="<YOUR_API_KEY>"
    PROFILE_ID = "youssef"

    ontbo = Ontbo(API_KEY)
    profile=ontbo.profile(PROFILE_ID)
    print(profile.query_facts("What is the user's job?"))


